import logging

from s3ben.constants import (
    DEFAULT_LOG_DATE_FORMAT,
    DEFAULT_LOG_FORMAT,
    DEFAULT_LOG_FORMAT_DEBUG,
)


def init_logger(
    name: str,
    level: str = "warning",
    log_format: str = None,
    date_format: str = DEFAULT_LOG_DATE_FORMAT,
) -> None:
    """
    Function to initialize logger and all needed parts

    :param str name: logger name to setup
    :param str level: Logging level
    :raises ValueError: If log level doens't exist
    :return: None
    """
    if not logging._checkLevel(level.upper()):
        raise ValueError(f"Log level {level} doesn't exist")
    set_level = logging.getLevelName(level.upper())
    if not log_format:
        log_format = (
            DEFAULT_LOG_FORMAT_DEBUG if level.lower() == "debug" else DEFAULT_LOG_FORMAT
        )
    set_format = logging.Formatter(log_format, datefmt=date_format)
    logger = logging.getLogger(name)
    logger.setLevel(set_level)
    console = logging.StreamHandler()
    console.setFormatter(set_format)
    console.setLevel(set_level)
    logger.addHandler(console)



import multiprocessing
import os
import pwd
from argparse import Namespace
from logging import getLogger

from s3ben.arguments import base_args
from s3ben.backup import BackupManager, ResolveRemmaping
from s3ben.config import parse_config
from s3ben.decorators import argument, command
from s3ben.logger import init_logger
from s3ben.rabbit import RabbitMQ
from s3ben.s3 import S3Events
from s3ben.sentry import init_sentry

_logger = getLogger(__name__)
args = base_args()
subparser = args.add_subparsers(dest="subcommand")


def main() -> None:
    """
    Entry point
    :raises ValueError: if config file not found
    :return: None
    """
    parsed_args = args.parse_args()
    if parsed_args.subcommand is None:
        args.print_help()
        return
    init_logger(name="s3ben", level=parsed_args.log_level)
    if os.path.isfile(parsed_args.sentry_conf):
        _logger.debug("Initializing sentry")
        init_sentry(config=parsed_args.sentry_conf)
    config = parse_config(parsed_args.config)
    parsed_args.func(config, parsed_args)


@command(parent=subparser)
def setup(config: dict, args: Namespace) -> None:
    """
    Cli command to add required cofiguration to s3 buckets and mq
    :param dict config: Parsed configuration dictionary
    :return: None
    """
    _logger.info("Checking backup root")
    main: dict = config.pop("s3ben")
    backup_root: str = main.pop("backup_root")
    user = pwd.getpwnam(main.pop("user"))
    if not os.path.exists(backup_root):
        os.mkdir(path=backup_root, mode=0o700)
        os.chown(path=backup_root, uid=user.pw_uid, gid=user.pw_gid)
    _logger.info("Setting up RabitMQ")
    mq_conf: dict = config.pop("amqp")
    exchange = mq_conf.pop("exchange")
    queue = mq_conf.pop("queue")
    routing_key = exchange
    mq_host = mq_conf.pop("host")
    mq_user = mq_conf.pop("user")
    mq_pass = mq_conf.pop("password")
    mq_port = mq_conf.pop("port")
    mq_virtualhost = mq_conf.pop("virtualhost")
    mq = RabbitMQ(
        hostname=mq_host,
        user=mq_user,
        password=mq_pass,
        port=mq_port,
        virtualhost=mq_virtualhost,
    )
    mq.prepare(exchange=exchange, queue=queue, routing_key=routing_key)
    _logger.info("Setting up S3")
    s3 = config.pop("s3")
    s3_events = S3Events(
        hostname=s3.pop("hostname"),
        access_key=s3.pop("access_key"),
        secret_key=s3.pop("secret_key"),
        secure=s3.pop("secure"),
    )
    all_buckets = s3_events.get_admin_buckets()
    exclude_buckets = s3.pop("exclude").split(",")
    exclude_buckets = [b.strip() for b in exclude_buckets]
    filtered_buckets = list(set(all_buckets) - set(exclude_buckets))
    s3_events.create_topic(
        mq_host=mq_host,
        mq_user=mq_user,
        mq_port=mq_port,
        mq_password=mq_pass,
        exchange=exchange,
        mq_virtualhost=mq_virtualhost,
    )
    for bucket in filtered_buckets:
        _logger.debug(f"Setting up bucket: {bucket}")
        s3_events.create_notification(bucket=bucket, exchange=exchange)


def init_consumer(config: dict) -> None:
    main = config.pop("s3ben")
    mq_conf: dict = config.pop("amqp")
    mq_host = mq_conf.pop("host")
    mq_user = mq_conf.pop("user")
    mq_pass = mq_conf.pop("password")
    mq_port = mq_conf.pop("port")
    queue = mq_conf.pop("queue")
    backup_root = main.pop("backup_root")
    s3 = config.pop("s3")
    s3_events = S3Events(
        hostname=s3.pop("hostname"),
        access_key=s3.pop("access_key"),
        secret_key=s3.pop("secret_key"),
        secure=s3.pop("secure"),
        backup_root=backup_root,
    )
    mq = RabbitMQ(
        hostname=mq_host,
        user=mq_user,
        password=mq_pass,
        port=mq_port,
        virtualhost=mq_conf.get("virtualhost"),
    )
    backup = BackupManager(
        backup_root=backup_root, user=main.pop("user"), mq=mq, mq_queue=queue
    )
    backup.start_consumer(s3_client=s3_events)


@command(parent=subparser)
def consume(config: dict, args: Namespace) -> None:
    s3ben_config: dict = config.get("s3ben")
    num_proc = s3ben_config.get("process") if "process" in s3ben_config.keys() else 8
    processes = []
    for _ in range(int(num_proc)):
        process = multiprocessing.Process(target=init_consumer, args=(config,))
        processes.append(process)
    for proc in processes:
        _logger.debug("Starting process: %s", proc)
        proc.start()
    for proc in processes:
        proc.join()


@command(
    [
        argument("--bucket", required=True, help="Bucket name which to sync", type=str),
        argument(
            "--transfers",
            help="Number of transfer processes, default: %(default)d",
            type=int,
            default=4,
        ),
        argument(
            "--checkers",
            help="Number of checker processes, default: %(default)d",
            type=int,
            default=4,
        ),
        argument("--skip-checksum", help="Skip checksum check", action="store_true"),
        argument("--skip-filesize", help="Skip filesize check", action="store_true"),
        argument(
            "--page-size",
            help="Bucket object page size, default: %(default)s",
            type=int,
            default=1000,
        ),
        argument(
            "--ui",
            help="Use experimental ui, default: %(default)s",
            action="store_true",
        ),
    ],
    parent=subparser,
)
def sync(config: dict, args: Namespace):
    """
    Entry point for sync cli option
    """
    _logger.debug("Initializing sync")
    s3 = config.pop("s3")
    backup_root = config["s3ben"].pop("backup_root")
    s3_events = S3Events(
        hostname=s3.pop("hostname"),
        access_key=s3.pop("access_key"),
        secret_key=s3.pop("secret_key"),
        secure=s3.pop("secure"),
        backup_root=backup_root,
    )
    backup = BackupManager(
        backup_root=backup_root,
        user=config["s3ben"].pop("user"),
        s3_client=s3_events,
        curses=args.ui,
    )
    backup.sync_bucket(
        args.bucket,
        args.transfers,
        args.page_size,
        args.checkers,
        args.skip_checksum,
        args.skip_filesize,
    )


@command(
    [
        argument("--show-excluded", help="Show excluded buckets", action="store_true"),
        argument("--show-obsolete", help="Show obsolete bucklets", action="store_true"),
        argument(
            "--only-enabled",
            help="Show only backup enabled buckets",
            action="store_true",
        ),
        argument(
            "--sort",
            help="Sort results by select collump, default: %(default)s",
            choices=["bucket", "owner", "size", "objects"],
            default="bucket",
        ),
        argument(
            "--sort-reverse", help="Reverse order for sorting", action="store_true"
        ),
    ],
    parent=subparser,
)
def buckets(config: dict, args: Namespace) -> None:
    _logger.debug("Listing buckets")
    s3 = config.pop("s3")
    exclude = s3.pop("exclude").replace('"', "").replace("'", "").split(",")
    exclude = [e.strip() for e in exclude]
    backup_root = config["s3ben"].pop("backup_root")
    s3_events = S3Events(
        hostname=s3.pop("hostname"),
        access_key=s3.pop("access_key"),
        secret_key=s3.pop("secret_key"),
        secure=s3.pop("secure"),
        backup_root=backup_root,
    )
    backup = BackupManager(
        backup_root=backup_root,
        user=config["s3ben"].pop("user"),
        s3_client=s3_events,
    )
    backup.list_buckets(
        exclude=exclude,
        show_excludes=args.show_excluded,
        show_obsolete=args.show_obsolete,
        only_enabled=args.only_enabled,
        sort=args.sort,
        sort_revers=args.sort_reverse,
    )


@command(
    [
        argument(
            "--days-keep",
            help="How long to keep, default: %(default)d",
            default=30,
            type=int,
        )
    ],
    parent=subparser,
)
def cleanup(config: dict, args: Namespace) -> None:
    """
    Cli function to call deleted items cleanup method
    from BackupManager
    """
    _logger.debug("Starting deleted items cleanup")
    backup_root = config["s3ben"].pop("backup_root")
    backup = BackupManager(
        backup_root=backup_root,
        user=config["s3ben"].pop("user"),
    )
    backup.cleanup_deleted_items(days=args.days_keep)


@command(parent=subparser)
def test(config: dict, args: Namespace) -> None:
    import multiprocessing
    import time

    backup_root = config["s3ben"].get("backup_root")
    queue_data = {
        "action": "update",
        "data": {"bucket": "test", "remap": {"somekey": "_alternate_/somekey"}},
    }
    remmaping = ResolveRemmaping(backup_root=backup_root)
    event = multiprocessing.Event()
    queue = multiprocessing.Queue()
    process = multiprocessing.Process(
        target=remmaping.run,
        args=(
            queue,
            event,
        ),
    )
    process.start()
    time.sleep(1.2)
    queue.put_nowait(queue_data)
    time.sleep(2)
    event.set()

    # remmaping.update_remapping("test2", remap)



DEFAULT_LOG_DATE_FORMAT = "%Y-%m-%d %H:%M:%S"
DEFAULT_LOG_FORMAT_DEBUG = "[%(asctime)s][%(levelname)-8s]: %(message)s"
DEFAULT_LOG_FORMAT = "%(message)s"
MANDATORY_CONFIG_KEYS = ["s3_endpoint", "aws_secret_key", "aws_access_key"]
TOPIC_ARN = "arn:aws:sns:default::{0}"
NOTIFICATION_EVENTS = [
    "s3:ObjectCreated:*",
    "s3:ObjectRemoved:*",
    "s3:ObjectLifecycle:Expiration:*",
    "s3:ObjectSynced:Create",
    "s3:ObjectSynced:Delete",
]
AMQP_HOST = "amqp://{user}:{password}@{host}:{port}/{virtualhost}"
UNITS = ["k", "M"]
SIZE_UNITS = ["", "Ki", "Mi", "Gi", "Ti", "Pi"]



import datetime
import itertools
import multiprocessing
import os
import shutil
import sys
from logging import getLogger
from typing import Union

import boto3
import botocore
import botocore.errorfactory
import rgwadmin
from rgwadmin import RGWAdmin

from s3ben.constants import AMQP_HOST, NOTIFICATION_EVENTS, TOPIC_ARN

_logger = getLogger(__name__)


class S3Events:
    """
    Class for configuring or showing config of the bucket
    :param str secret_key: Secret key fro s3
    :param str access_key: Access key for s3
    :param str endpoint: S3 endpoint uri
    """

    def __init__(
        self,
        secret_key: str,
        access_key: str,
        hostname: str,
        secure: bool,
        backup_root: str = None,
    ) -> None:
        self._download = os.path.join(backup_root, "active") if backup_root else None
        self._remove = os.path.join(backup_root, "deleted") if backup_root else None
        protocol = "https" if secure else "http"
        endpoint = f"{protocol}://{hostname}"
        self.client_s3 = boto3.client(
            service_name="s3",
            region_name="default",
            endpoint_url=endpoint,
            aws_access_key_id=access_key,
            aws_secret_access_key=secret_key,
        )
        self.client_sns = boto3.client(
            service_name="sns",
            region_name="default",
            endpoint_url=endpoint,
            aws_access_key_id=access_key,
            aws_secret_access_key=secret_key,
            config=botocore.client.Config(signature_version="s3"),
        )
        self.client_admin = RGWAdmin(
            access_key=access_key, secret_key=secret_key, server=hostname, secure=secure
        )
        self.session = boto3.Session(
            region_name="default",
            aws_access_key_id=access_key,
            aws_secret_access_key=secret_key,
        )
        self.resouce = self.session.resource(service_name="s3", endpoint_url=endpoint)

    def get_config(self, bucket: str):
        return self.client_s3.get_bucket_notification_configuration(Bucket=bucket)

    def create_bucket(self, bucket: str) -> None:
        """
        Create empty bucket with no configuration
        :param str bucket: Bucket name to create
        :return: None
        """
        self.client_s3.create_bucket(Bucket=bucket)

    def create_topic(
        self,
        mq_host: str,
        mq_user: str,
        mq_password: str,
        exchange: str,
        mq_port: int,
        mq_virtualhost: str,
    ) -> None:
        """
        Create bucket event notification config
        :param str bucket: Bucket name for config update
        :param str amqp: rabbitmq address
        """
        amqp = AMQP_HOST.format(
            user=mq_user,
            password=mq_password,
            host=mq_host,
            port=mq_port,
            virtualhost=mq_virtualhost,
        )
        attributes = {
            "push-endpoint": amqp,
            "amqp-exchange": exchange,
            "amqp-ack-level": "broker",
            "persistent": "true",
        }
        self.client_sns.create_topic(Name=exchange, Attributes=attributes)

    def create_notification(self, bucket: str, exchange: str) -> None:
        """
        Create buclet notification config
        :param str bucket: Bucket name
        :param str exchange: Exchange name were to send notification
        """
        notification_config = {
            "TopicConfigurations": [
                {
                    "Id": f"s3ben-{exchange}",
                    "TopicArn": TOPIC_ARN.format(exchange),
                    "Events": NOTIFICATION_EVENTS,
                }
            ]
        }
        self.client_s3.put_bucket_notification_configuration(
            Bucket=bucket, NotificationConfiguration=notification_config
        )

    def get_admin_buckets(self) -> list:
        """
        Admin api get buckets
        :return: list
        """
        return self.client_admin.get_buckets()

    def get_bucket(self, bucket: str) -> dict:
        """
        Get bucket info via admin api
        :param str bucket: Bucket name to fetch info
        :return: dictionary with bucket info
        """
        try:
            return self.client_admin.get_bucket(bucket=bucket)
        except rgwadmin.exceptions.NoSuchBucket:
            _logger.warning(f"Bucket {bucket} not found")
            sys.exit()

    def __decuple_download(self, input: tuple) -> None:
        bucket, path = input
        self.download_object(bucket, path)

    def download_object(self, bucket: str, path: Union[str, dict]):
        """
        Get an object from a bucket

        :param str bucket: Bucket name from which to get object
        :param str path: object path
        """

        s3_obj = path
        if isinstance(path, dict):
            dst = next(iter(path.values()))
            s3_obj = "/" + next(iter(path.keys()))
            destination = os.path.join(self._download, bucket, dst)
        else:
            destination = os.path.join(self._download, bucket, path)
        dst_dir = os.path.dirname(destination)
        if not os.path.exists(dst_dir):
            os.makedirs(dst_dir, exist_ok=True)
        try:
            self.client_s3.head_object(Bucket=bucket, Key=s3_obj)
        except botocore.exceptions.ClientError as err:
            if err.response["ResponseMetadata"]["HTTPStatusCode"] == 404:
                _logger.warning("%s not found in bucket: %s", path, bucket)
        _logger.info("Downloading: %s:%s to %s", bucket, s3_obj, destination)
        self.client_s3.download_file(Bucket=bucket, Key=s3_obj, Filename=destination)

    def remove_object(self, bucket: str, path: str) -> None:
        """
        Move object to deleted items
        :param str bucket: Bucket eame
        :param str path: object path which should be moved
        :return: None
        """
        _logger.info(f"Moving {path} to deleted items for bucket: {bucket}")
        current_date = datetime.date.today().strftime("%Y-%m-%d")
        dest = os.path.dirname(os.path.join(self._remove, current_date, bucket, path))
        src = os.path.join(self._download, bucket, path)
        file_name = os.path.basename(path)
        d_file = os.path.join(dest, file_name)
        if not os.path.exists(src):
            _logger.warning(f"{src} doesn't exist")
            return
        if not os.path.exists(dest):
            os.makedirs(dest)
        if os.path.isfile(d_file):
            _logger.warning(
                f"Removing {d_file} as another with same name must be moved to deleted items"
            )
            os.remove(d_file)
        shutil.move(src, dest)

    def download_all_objects(self, bucket_name: str, obj_keys: list) -> None:
        """
        Method for getting all objects from one bucket
        :param str bucket_name: Name of the bucket
        :param str dest: Directory root to append
        :param int threads: Number of threads to start
        :return: None
        """
        threads = 2
        with multiprocessing.pool.ThreadPool(threads) as threads:
            iterate = zip(itertools.repeat(bucket_name), obj_keys)
            threads.map(self.__decuple_download, iterate)

    def _get_all_objects(self, bucket_name) -> list:
        """
        Method to get all objects from the bucket
        :param str bucket_name: Name of the bucket
        :return: List all objects in the bucket
        """
        objects = self.resouce.Bucket(bucket_name).objects.all()
        return [o.key for o in objects]

    def download_all_objects_v2(
        self, baucket_name: str, step: int, page_queue: multiprocessing.Queue
    ) -> None:
        """
        Download objects from page object iterating every n'th page
        :param str bucket_name: Name of the bucket
        :param page: bot paginator
        :param int step: For loop step for multiprocess support
        :param multiprocessing.Queue queu: Multiprocess queu for exchanging information
        """
        import time

        print("Process started")
        while True:
            data = page_queue.get(block=True, timeout=10)
            print(len(data[0]["Contents"]))
            if page_queue.empty():
                print("breaking")
                break
            time.sleep(0.5)

    def _check_if_object_exists(self, path):
        pass



import grp
import math
import os
import pwd
import sys
import time
from datetime import datetime

from s3ben.constants import SIZE_UNITS, UNITS


def drop_privileges(user: str) -> None:
    """
    Drop user privileges.

    :params str user: Username to which we should change permissions
    """
    new_user = pwd.getpwnam(user)
    if new_user.pw_uid == os.getuid():
        return
    new_gids = [new_user.pw_gid]
    new_gids += [
        group.gr_gid for group in grp.getgrall() if new_user.pw_name in group.gr_mem
    ]
    os.setgroups(new_gids[: os.NGROUPS_MAX])
    os.setgid(new_user[0])
    os.setuid(new_user.pw_uid)
    os.environ["HOME"] = new_user.pw_dir


def convert_to_human_v2(value: int):
    """
    Convert size to human
    """
    suffix = "B"
    for unit in SIZE_UNITS:
        if abs(value) < 1024.0:
            break
        if unit == UNITS[-1]:
            break
        value /= 1024.0
    return f"{value:3.2f}{unit}{suffix}"


def convert_to_human(value: int) -> tuple:
    if float(value) <= 1000.0:
        return value, ""
    for unit in UNITS:
        value /= 1000.0
        if float(value) < 1000.0:
            break
        if unit == UNITS[-1]:
            break
    return value, unit


class ProgressBar:
    """
    Progress bar class
    """

    _skipped: int = 0
    _percents = "[{0:>6.2f}%]"
    _time_left = "[LEFT: {0:0>2}:{1:0>2}:{2:0>2}]"
    _running = "[RUN: {0:0>2}:{1:0>2}:{2:0>2}]"
    _progress = "[{:{done_marker}>{done_size}}{}{:{base_marker}>{left_size}}]"
    _completed: float = 0
    _download: float = 0
    current_marker: list = ["-", "\\", "|", "/"]
    filler_marker: str = "."
    bar_length: int = 0
    bar_size: int = 0
    _preffix: str = ""
    _suffix: str = ""
    _done_marker: str = "█"
    _total: float = 100.00

    def __init__(
        self,
    ):
        self.terminal_size: int = os.get_terminal_size().columns
        self.percents: str = self._percents.format(0)
        self.show_numbers: bool = False
        self.time_start = int(time.perf_counter())
        self.time_left = self._time_left.format(99, 59, 59)
        self.run_time = self._running.format(0, 0, 0)
        self._run_time = int(time.perf_counter())
        self.avg_speed = 0

    def __update_stats(self) -> None:
        self.__run_time()
        self.__update_avg_speed()
        self.__update_percent_done()
        self.__calculate_estimate()
        self.__update_run_time()
        self.__update_terminal_size()

    def __get_current_marker(self) -> str:
        marker = self.current_marker.pop(0)
        self.current_marker.append(marker)
        return self.current_marker[-1]

    def __run_time(self) -> None:
        self._run_time = int(time.perf_counter()) - self.time_start

    def __update_terminal_size(self) -> None:
        self.terminal_size = os.get_terminal_size().columns

    def __split_time(self, seconds: int) -> tuple:
        hours = math.floor(seconds / 3600)
        minutes = math.floor(seconds / 60) - (hours * 60)
        seconds = math.floor(seconds - ((hours * 3600) + (minutes * 60)))
        return (hours, minutes, seconds)

    def __update_run_time(self) -> None:
        r_time = self.__split_time(self._run_time)
        self.run_time = self._running.format(r_time[0], r_time[1], r_time[2])

    def __update_avg_speed(self) -> None:
        try:
            self.avg_speed = self.progress / self._run_time
        except ZeroDivisionError:
            self.avg_speed = 0

    def __calculate_estimate(self) -> None:
        run_data_left = self.total - self.progress
        try:
            run_estimate = run_data_left // self.avg_speed
            e_time = self.__split_time(run_estimate)
            self.time_left = self._time_left.format(e_time[0], e_time[1], e_time[2])
        except ZeroDivisionError:
            self.time_left = self._time_left.format(99, 59, 59)

    def __update_percent_done(self) -> None:
        percents = float(self.progress * 100 / self.total)
        self.percents = self._percents.format(percents)

    def __format_skipped(self) -> str:
        """
        Format skipped progress bar

        :return: String representing skipped files part
        """

        skipped, s_units = convert_to_human(self._skipped)
        _skip_format_int = "[S:{0:>7}"
        _skip_format_float = "[S:{0:>6.2f}{1:1}"

        if self._skipped <= 1000:
            return _skip_format_int.format(skipped)
        return _skip_format_float.format(skipped, s_units)

    def __format_downloaded(self) -> str:
        """
        Method to format download files

        :return: String representing download part of progress
        """

        downloaded, d_units = convert_to_human(self._download)
        _download_format_int = "|DL:{0:7}"
        _download_format_float = "|DL:{0:6.2f}{1:1}"

        if self._download <= 1000:
            return _download_format_int.format(downloaded)
        return _download_format_float.format(downloaded, d_units)

    def __format_total(self) -> str:
        """
        Format total part for pregress bar
        :returns: String
        """
        results = []
        total, t_units = convert_to_human(self.total)
        progress, units = convert_to_human(self.progress)
        _progress_format_int = "|{0:>7}"
        _progress_format_float = "|{0:>6.2f}{1:<1}"
        _total_format_int = "/{0:<4}]"
        _total_format_float = "/{0:<.2f}{1:<1}]"
        if self.progress <= 1000:
            results.append(_progress_format_int.format(progress))
        else:
            results.append(_progress_format_float.format(progress, units))
        if self._total <= 1000:
            results.append(_total_format_int.format(total))
        else:
            results.append(_total_format_float.format(total, t_units))
        return "".join(results)

    def __format_bar(self) -> str:
        p_bar = [""]
        bar_info = 0
        bar_index = 0
        finished = int(self.bar_size * self.progress / self.total)
        skipped = self.__format_skipped()
        download = self.__format_downloaded()
        total = self.__format_total()
        p_bar.append(skipped)
        bar_info += len(skipped)
        p_bar.append(download)
        bar_info += len(download)
        p_bar.append(total)
        bar_info += len(total)
        p_bar.append("")
        bar_index = len(p_bar) - 1
        p_bar.append(self.percents)
        bar_info += len(self.percents)
        p_bar.append(self.run_time)
        bar_info += len(self.run_time)
        p_bar.append(self.time_left)
        bar_info += len(self.time_left)
        if bar_index:
            if self.terminal_size > bar_info:
                self.bar_size = self.terminal_size - bar_info - 3
                left_size = self.bar_size - finished
                p_bar[bar_index] = self._progress.format(
                    self._done_marker if finished > 1 else "",
                    self.__get_current_marker() if left_size > 0 else "",
                    self.filler_marker if left_size > 0 else "",
                    done_marker=self._done_marker,
                    done_size=finished if left_size > 0 else self.bar_size,
                    in_progress_marker=self.filler_marker,
                    base_marker=self.filler_marker,
                    left_size=left_size if left_size > 0 else 0,
                )
            else:
                p_bar.pop(bar_index)
        if self.bar_size <= finished:
            self.current_marker = ""
        p_bar.append("\r")
        line = "".join(p_bar)
        return line

    @property
    def progress(self) -> float:
        return self._completed

    @progress.setter
    def progress(self, data: dict) -> None:
        """
        Update progress done percentage
        """
        if "skipped" in data.keys():
            skipped = data.get("skipped")
            self._skipped += skipped
            self._completed += skipped
        if "downloaded" in data.keys():
            download = data.get("downloaded")
            self._download += download
            self._completed += download

    @property
    def total(self):
        """
        Return total value of progress calculation
        """
        return self._total

    @total.setter
    def total(self, total) -> None:
        """
        Set progress bar total value
        """
        self._total = total

    def draw(self) -> None:
        """
        Draw a progress bar
        """
        self.__update_stats()
        sys.stdout.write(self.__format_bar())
        sys.stdout.flush()

    def __del__(self) -> None:
        sys.stdout.write("\n")
        sys.stdout.flush()



import functools
import json
from logging import getLogger

import pika

from s3ben.s3 import S3Events

_logger = getLogger(__name__)


class RabbitMQ:
    """
    Class to setup and consume rabbitmq cluster
    :param str hostname: rabbitmq server address
    :param str user: username for connection
    :param str password: password for user
    :param int port: port of mq server, default: 5672
    :param str virtualhost: virtual host to connect, default: /
    """

    def __init__(
        self, hostname: str, user: str, password: str, port: int, virtualhost: str
    ) -> None:
        mq_credentials = pika.credentials.PlainCredentials(user, password)
        self.mq_params = pika.ConnectionParameters(
            host=hostname,
            port=port,
            virtual_host=virtualhost,
            credentials=mq_credentials,
        )
        self.should_reconnect: bool = False
        self.was_consuming = False

        self._connection: pika.SelectConnection = None
        self._channel: pika.channel.Channel = None
        self._consuming: bool = False
        self._closing: bool = False
        self._consumer_tag = None
        self._prefetch_count = 1
        self._s3_client: S3Events = None
        self._exchange: str = None
        self._queue: str = None
        self._routing_key: str = None

    def prepare(self, exchange: str, queue: str, routing_key: str) -> None:
        """
        Method to prepare rabbitmq cluster for consumer
        :param str exchange: Exchange to create
        :param str queue: queue to create
        :param str routing_key: bind queu to exchange ussing routing key
        """
        _logger.info("Preparing rabbitmq cluster for work")
        connection = pika.BlockingConnection(parameters=self.mq_params)
        self._channel = connection.channel()
        self.setup_exchange(exchange=exchange)
        self.setup_queue(queue=queue)
        self.quene_bind(routing_key=routing_key)
        self.close_channel()
        connection.close()

    def consume(self, queue: str, s3_client: S3Events) -> None:
        """
        This method connects to RabbitMQ, returning the connection handle.
        When the connection is established, the on_connection_open method
        will be invoked by pika.

        :rtype: pika.SelectConnection
        """
        _logger.info("Connecting to MQ")
        self._s3_client = s3_client
        self._queue = queue
        self._connection = pika.SelectConnection(
            parameters=self.mq_params,
            on_open_callback=self.on_connection_open,
            on_open_error_callback=self.on_connection_open_error,
            on_close_callback=self.on_connection_closed,
        )
        self._connection.ioloop.start()

    def on_connection_open(self, _unused_connection) -> None:
        """This method is called by pika once the connection to RabbitMQ has
        been established. It passes the handle to the connection object in
        case we need it, but in this case, we'll just mark it unused.

        :param pika.SelectConnection _unused_connection: The connection
        """
        _logger.debug("Connection opened")
        self.open_channel()

    def open_channel(self) -> None:
        """
        Open a new channel with RabbitMQ by issuing the Channel.Open RPC
        command. When RabbitMQ responds that the channel is open, the
        on_channel_open callback will be invoked by pika.
        """
        _logger.debug("Creating new channel")
        self._connection.channel(on_open_callback=self.on_channel_open)

    def on_channel_open(self, channel: pika.channel.Channel) -> None:
        """
        This method is invoked by pika when the channel has been opened.
        The channel object is passed in so we can make use of it.
        Since the channel is now open, we'll declare the exchange to use.

        :param pika.channel.Channel channel: The channel object
        """
        _logger.debug("Channel opened")
        self._channel = channel
        self.add_on_channel_close_callback()
        self.set_qos()

    def add_on_channel_close_callback(self) -> None:
        """
        This method tells pika to call the on_channel_closed method if
        RabbitMQ unexpectedly closes the channel.
        """
        _logger.debug("Adding channel close callback")
        self._channel.add_on_close_callback(self.on_channel_closed)

    def on_channle_closed(self, channel, reason) -> None:
        """
        Invoked by pika when RabbitMQ unexpectedly closes the channel.
        Channels are usually closed if you attempt to do something that
        violates the protocol, such as re-declare an exchange or queue with
        different parameters. In this case, we'll close the connection
        to shutdown the object.

        :param pika.channel.Channel: The closed channel
        :param Exception reason: why the channel was closed
        """
        _logger.warning(f"Channel {channel} was closed: {reason}")
        self.close_connection()

    def close_connection(self) -> None:
        """
        Close connection

        :return: None
        """
        self._consuming = False
        if self._connection.is_closing or self._connection.is_closed:
            _logger.info("Connection closing or already closed")
            return
        _logger.info("Closing connection")
        self._connection.close()

    def on_connection_open_error(self, _unused_connection, err) -> None:
        """
        This method is called by pika if the connection to RabbitMQ
        can't be established.

        :param pika.SelectConnection _unused_connection: The connection
        :param Exception err: The error
        """
        _logger.error(f"Connection open failed: {err}")
        self.reconnect()

    def reconnect(self) -> None:
        """
        Will be invoked if the connection can't be opened or is
        closed. Indicates that a reconnect is necessary then stops the
        ioloop.
        """
        self.should_reconnect = True
        self.stop()

    def stop(self) -> None:
        """
        Cleanly shutdown the connection to RabbitMQ by stopping the consumer
        with RabbitMQ. When RabbitMQ confirms the cancellation, on_cancelok
        will be invoked by pika, which will then closing the channel and
        connection. The IOLoop is started again because this method is invoked
        when CTRL-C is pressed raising a KeyboardInterrupt exception. This
        exception stops the IOLoop which needs to be running for pika to
        communicate with RabbitMQ. All of the commands issued prior to starting
        the IOLoop will be buffered but not processed.
        """
        if not self._closing:
            self._closing = True
            _logger.info("Stopping")
            if self._consuming:
                _logger.debug("still consuming")
                self.stop_consuming()
                self._connection.ioloop.start()
            else:
                self._connection.ioloop.stop()
            _logger.info("Stopped")

    def stop_consuming(self) -> None:
        """
        Tell RabbitMQ that you would like to stop consuming by sending the
        Basic.Cancel RPC command.
        """
        if self._channel:
            _logger.debug("Sending a Basic.Cancel RPC command to RabbitMQ")
            callback = functools.partial(self.on_cancelok, userdata=self._consumer_tag)
            self._channel.basic_cancel(self._consumer_tag, callback)

    def on_cancelok(self, _unused_frame, userdata) -> None:
        """
        This method is invoked by pika when RabbitMQ acknowledges the
        cancellation of a consumer. At this point we will close the channel.
        This will invoke the on_channel_closed method once the channel has been
        closed, which will in-turn close the connection.

        :param pika.frame.Method _unused_frame: The Basic.CancelOk frame
        :param str|unicode userdata: Extra user data (consumer tag)
        """
        self._consuming = False
        _logger.debug(
            f"RabbitMQ acknowledged the cancellation of the consumer: {userdata}"
        )
        self.close_channel()

    def close_channel(self) -> None:
        """
        Call to close the channel with RabbitMQ cleanly by issuing the
        Channel.Close RPC command.
        """
        _logger.info("Closing the channel")
        self._channel.close()

    def on_connection_closed(self, _unused_connection, reason) -> None:
        """
        This method is invoked by pika when the connection to RabbitMQ is
        closed unexpectedly. Since it is unexpected, we will reconnect to
        RabbitMQ if it disconnects.

        :param pika.connection.Connection connection: The closed connection obj
        :param Exception reason: exception representing reason for loss of
            connection.
        """
        self._channel = None
        if self._closing:
            self._connection.ioloop.stop()
            return
        _logger.warning(f"Connection closed, reconnecting because: {reason}")
        self.reconnect()

    def on_channel_closed(self, channel, reason) -> None:
        """
        Invoked by pika when RabbitMQ unexpectedly closes the channel.
        Channels are usually closed if you attempt to do something that
        violates the protocol, such as re-declare an exchange or queue with
        different parameters. In this case, we'll close the connection
        to shutdown the object.

        :param pika.channel.Channel: The closed channel
        :param Exception reason: why the channel was closed
        """
        _logger.warning(f"Channel {channel} was closed: {reason}")
        self.close_connection()

    def setup_queue(self, queue) -> None:
        """
        Setup the queue on RabbitMQ by invoking the Queue.Declare RPC
        command.

        :param str|unicode queue_name: The name of the queue to declare.
        """
        _logger.debug(f"Creating queue: {queue}")
        self._queue = queue
        self._channel.queue_declare(
            queue=self._queue,
            durable=True,
            auto_delete=False,
            arguments={"x-queue-type": "quorum"},
        )

    def quene_bind(self, routing_key: str) -> None:
        _logger.debug(f"Binding {self._exchange} to {self._queue} with {routing_key}")
        self._channel.queue_bind(
            queue=self._queue, exchange=self._exchange, routing_key=routing_key
        )

    def setup_exchange(self, exchange) -> None:
        """
        Setup the exchange on RabbitMQ by invoking the Exchange.Declare RPC
        command.

        :param str|unicode exchange_name: The name of the exchange to declare
        """
        _logger.debug(f"Declaring exhange {exchange}")
        self._exchange = exchange
        self._channel.exchange_declare(
            exchange=self._exchange,
            durable=True,
            exchange_type="direct",
            auto_delete=False,
            internal=False,
        )

    def start_consuming(self) -> None:
        """This method sets up the consumer by first calling
        add_on_cancel_callback so that the object is notified if RabbitMQ
        cancels the consumer. It then issues the Basic.Consume RPC command
        which returns the consumer tag that is used to uniquely identify the
        consumer with RabbitMQ. We keep the value to use it when we want to
        cancel consuming. The on_message method is passed in as a callback pika
        will invoke when a message is fully received.
        """
        _logger.info("Starting to consume messages")
        self.add_on_cancel_callback()
        self._consumer_tag = self._channel.basic_consume(self._queue, self.on_message)
        self.was_consuming = True
        self._consuming = True

    def add_on_cancel_callback(self) -> None:
        """
        Add a callback that will be invoked if RabbitMQ cancels the consumer
        for some reason. If RabbitMQ does cancel the consumer,
        on_consumer_cancelled will be invoked by pika.
        """
        _logger.debug("Adding consumer cancel callback")
        self._channel.add_on_cancel_callback(self.on_consumer_cancelled)

    def on_consumer_cancelled(self, method_frame) -> None:
        """
        Invoked by pika when RabbitMQ sends a Basic.Cancel for a consumer
        receiving messages.

        :param pika.frame.Method method_frame: The Basic.Cancel frame
        """
        _logger.info(f"Consumer canceled, shuting down {method_frame}")
        if self._channel:
            self._channel.close()

    def on_message(self, _unused_channel, basic_deliver, properties, body) -> None:
        """
        Invoked by pika when a message is delivered from RabbitMQ. The
        channel is passed for your convenience. The basic_deliver object that
        is passed in carries the exchange, routing key, delivery tag and
        a redelivered flag for the message. The properties passed in is an
        instance of BasicProperties with the message properties and the body
        is the message that was sent.

        :param pika.channel.Channel _unused_channel: The channel object
        :param pika.Spec.Basic.Deliver: basic_deliver method
        :param pika.Spec.BasicProperties: properties
        :param bytes body: The message body
        """
        body = json.loads(body.decode())
        for record in body["Records"]:
            obj_size = record["s3"]["object"]["size"]
            if obj_size == 0:
                continue
            event = record["eventName"]
            bucket = record["s3"]["bucket"]["name"]
            obj_key = record["s3"]["object"]["key"]
            _logger.debug(
                f"Bucket: {bucket} event: {event} file: {obj_key} size: {obj_size}"
            )
            action = self.__bucket_event(event=event)
            if action == "download":
                self._s3_client.download_object(bucket=bucket, path=obj_key)
            if action == "remove":
                self._s3_client.remove_object(bucket=bucket, path=obj_key)
        self.acknowledge_message(basic_deliver.delivery_tag)

    def __bucket_event(self, event: str) -> str:
        """
        Method to parse bucket event
        :param str event: Event that was received
        :raises ValueError: if event unknow
        :return: Parsed action to execute
        """
        e_split = event.split(":")
        if e_split[0].lower() == "objectsynced" and e_split[1].lower() == "create":
            return "download"
        if e_split[0].lower() == "objectsynced" and e_split[1].lower() == "delete":
            return "remove"
        if e_split[0].lower() == "objectcreated":
            return "download"
        if e_split[0].lower() == "objectremoved":
            return "remove"
        if e_split[0].lower() == "objectlifecycle":
            return "remove"
        raise ValueError(f"Event: {event} unknow")

    def acknowledge_message(self, delivery_tag) -> None:
        """
        Acknowledge the message delivery from RabbitMQ by sending a
        Basic.Ack RPC method for the delivery tag.

        :param int delivery_tag: The delivery tag from the Basic.Deliver frame
        """
        _logger.debug(f"Ack message: {delivery_tag}")
        self._channel.basic_ack(delivery_tag)

    def on_basic_qos_ok(self, _unused_frame):
        """Invoked by pika when the Basic.QoS method has completed. At this
        point we will start consuming messages by calling start_consuming
        which will invoke the needed RPC commands to start the process.

        :param pika.frame.Method _unused_frame: The Basic.QosOk response frame

        """
        _logger.info("QOS set to: %d", self._prefetch_count)
        self.start_consuming()

    def set_qos(self):
        """This method sets up the consumer prefetch to only be delivered
        one message at a time. The consumer must acknowledge this message
        before RabbitMQ will deliver another one. You should experiment
        with different prefetch values to achieve desired performance.

        """
        _logger.debug(f"Setting prefetch: {self._prefetch_count}")
        self._channel.basic_qos(
            prefetch_count=self._prefetch_count, callback=self.on_basic_qos_ok
        )



import configparser
import os
from logging import getLogger
from pathlib import Path

_logger = getLogger(__name__)


def parse_config(file: Path) -> dict:
    """
    Function to parse base config file
    :param Path file: Path to main config file
    :return: RawConfigParser
    """
    if not os.path.exists(file):
        raise ValueError(f"Config {file} not found")
    _logger.debug(f"Parsing {file}")
    config = configparser.RawConfigParser()
    with open(file=file) as f:
        config.readfp(f)
    results = dict(config.__dict__["_sections"])
    return results



import hashlib
import json
import multiprocessing
import os
import shutil
import signal
import time
from datetime import datetime, timedelta
from logging import getLogger
from multiprocessing import Event, Queue
from queue import Empty
from typing import Union

from tabulate import tabulate

from s3ben.helpers import (
    ProgressBar,
    convert_to_human,
    convert_to_human_v2,
    drop_privileges,
)
from s3ben.rabbit import RabbitMQ
from s3ben.s3 import S3Events
from s3ben.ui import S3benGui

_logger = getLogger(__name__)


class BackupManager:
    """
    Class to coordinate all tasks

    :param str backup_root: Destination directory were all files will be placed
    :param str user: username to which change privileges
    :param str mq_queue: rabbitmq queue name
    :param RabbitMQ mq: RabbitMQ class object
    """

    def __init__(
        self,
        backup_root: str,
        user: str,
        mq_queue: str = None,
        mq: RabbitMQ = None,
        s3_client: S3Events = None,
        curses: bool = False,
    ):
        self._backup_root = backup_root
        self._user = user
        self._mq = mq
        self._mq_queue = mq_queue
        self._s3_client = s3_client
        self._bucket_name: str = None
        self._page_size: int = None
        self._progress_queue = None
        self._download_queue = None
        self._verify_queue = None
        self._remapped_queue = None
        self._end_event = None
        self._barrier = None
        self._curses = curses
        signal.signal(signal.SIGTERM, self.__exit)
        signal.signal(signal.SIGINT, self.__exit)

    def __exit(self, signal_no, stack_frame) -> None:
        raise SystemExit()

    def start_consumer(self, s3_client: S3Events) -> None:
        _logger.debug(f"Dropping privileges to {self._user}")
        drop_privileges(user=self._user)
        try:
            self._mq.consume(queue=self._mq_queue, s3_client=s3_client)
        except KeyboardInterrupt:
            self._mq.stop()
        except SystemExit:
            self._mq.stop()

    def _progress(self) -> None:
        progress = ProgressBar()
        info = self._s3_client.get_bucket(self._bucket_name)
        total_objects = info["usage"]["rgw.main"]["num_objects"]
        progress.total = total_objects
        progress.draw()
        while progress.total > progress.progress:
            try:
                data = self._progress_queue.get(timeout=0.5)
            except Empty:
                progress.draw()
                continue
            else:
                progress.progress = data
                progress.draw()

    def sync_bucket(
        self,
        bucket_name: str,
        threads: int,
        page_size: int,
        checkers: int,
        skip_checksum: bool,
        skip_filesize: bool,
    ) -> None:
        _logger.info("Starting bucket sync")
        start = time.perf_counter()
        self._page_size = page_size
        self._bucket_name = bucket_name
        proc_manager = multiprocessing.managers.SyncManager()
        proc_manager.start()
        self._download_queue = proc_manager.Queue(maxsize=threads * 2)
        self._verify_queue = proc_manager.Queue(maxsize=threads * 2)
        self._progress_queue = proc_manager.Queue()
        self._remapped_queue = proc_manager.Queue()
        self._end_event = proc_manager.Event()
        self._barrier = proc_manager.Barrier(threads + checkers)
        try:
            reader = multiprocessing.Process(target=self._page_reader)
            reader.start()
            remapper = multiprocessing.Process(target=self.__remapper)
            remapper.start()
            processess = []
            for _ in range(checkers):
                verify = multiprocessing.Process(
                    target=self._page_verfication,
                    args=(
                        skip_checksum,
                        skip_filesize,
                    ),
                )
                processess.append(verify)
            for _ in range(threads):
                download = multiprocessing.Process(target=self._page_processor)
                processess.append(download)
            for proc in processess:
                proc.start()
            processess.append(reader)
            processess.append(remapper)
            if not self._curses:
                self._progress()
            else:
                self._curses_ui()
            for proc in processess:
                proc.join()
        finally:
            proc_manager.shutdown()
        end = time.perf_counter()
        _logger.info(f"Sync took: {round(end - start, 2)} seconds")

    def _curses_ui(self) -> None:
        info = self._s3_client.get_bucket(self._bucket_name)
        total_objects = info["usage"]["rgw.main"]["num_objects"]
        ui = S3benGui(title=f"Syncing buclet: {self._bucket_name}", total=total_objects)
        while True:
            try:
                data = self._progress_queue.get(timeout=0.5)
            except Empty:
                ui.progress(progress=ui._progress)
                continue
            else:
                ui.progress(progress=ui._progress + data)

    def __remapper(self) -> None:
        """
        Method to run remmaper class
        :return: None
        """
        remap_resolver = ResolveRemmaping(backup_root=self._backup_root)
        remap_resolver.run(queue=self._remapped_queue, event=self._end_event)

    def _page_reader(self) -> None:
        _logger.info("Starting page processing")
        self._end_event.clear()
        paginator = self._s3_client.client_s3.get_paginator("list_objects_v2")
        page_config = {"PageSize": self._page_size}
        pages = paginator.paginate(
            Bucket=self._bucket_name, PaginationConfig=page_config
        )
        for page in pages:
            self._verify_queue.put(page["Contents"])
        _logger.debug("Finished reading pages")
        self._end_event.set()

    def __check_object(self, path) -> Union[str, dict]:
        """
        Check object path for forward slash and replace
        if found as first character
        :param str path: Object path from s3
        """
        if path[0] == "/":
            _logger.error("Forward slash found as first simbol: %s", path)
            return {path[1:]: "_forward_slash_" + path}
        return path

    def _page_verfication(self, skip_checksum: bool, skip_filesize: bool) -> None:
        """
        Method to verify file by comparing size and checksum
        :returns: None
        """
        _logger.debug("Running page verification")
        self._barrier.wait()
        while True:
            try:
                data = self._verify_queue.get(block=False)
            except Empty:
                if self._end_event.is_set():
                    _logger.debug("Braking from queue")
                    break
                continue
            download_list = []
            for obj in data:
                obj_key = obj.get("Key")
                key = self.__check_object(obj_key)
                if isinstance(key, dict):
                    remapping_update = {"action": "update"}
                    local_path = next(iter(key.values()))
                    remapping_update.update(
                        {"data": {"bucket": self._bucket_name, "remap": key}}
                    )
                    self._remapped_queue.put(remapping_update)
                else:
                    local_path = key
                fp_key = os.path.join(
                    self._backup_root, "active", self._bucket_name, local_path
                )
                ob_key_exists = self.__check_file(path=fp_key)
                if not ob_key_exists:
                    _logger.debug("file doesn't exists: %s", fp_key)
                    download_list.append(key)
                    continue
                if not skip_checksum:
                    print("checking checksum")
                    obj_sum = obj.get("ETag")
                    obj_sum_matches = self.__check_md5(path=fp_key, md5=obj_sum)
                    if not obj_sum_matches:
                        download_list.append(key)
                        continue
                if not skip_filesize:
                    print("checking file size")
                    obj_size = obj.get("Size")
                    obj_size_matches = self.__check_file_size(
                        path=fp_key, size=obj_size
                    )
                    if not obj_size_matches:
                        download_list.append(key)
                        continue
            skipped = len(data) - len(download_list)
            progress_update = {"skipped": skipped}
            self._progress_queue.put(progress_update)
            if len(download_list) > 0:
                self._download_queue.put(download_list)

    def __check_file_size(self, path: str, size: int) -> bool:
        """
        Method to check if file size matches
        :param str path: full path to the file
        :param int size: size of remote object to verify
        :return: True if matches, otherwise False
        """
        _logger.debug(f"Checking file size {path}")
        local_size = os.stat(path=path).st_size
        if local_size == size:
            return True
        return False

    def __check_file(self, path: str) -> bool:
        """
        Method to check if local file exists
        :return: True if file exisrts, otherwise false
        """
        _logger.debug(f"Checking if file exists: {path}")
        if os.path.isfile(path=path):
            return True
        return False

    def __check_md5(self, path: str, md5: str) -> bool:
        """
        Method to calculate file md5 sum and check if it matches
        :param str path: full path to the local file
        :param str md5: md5 to check against
        :return: True if sum maches, otherwise false
        """
        _logger.debug(f"Checking md5sum for {path}")
        with open(path, "rb") as file:
            source_md5 = hashlib.md5()
            while chunk := file.read(8192):
                source_md5.update(chunk)
        calculated_md5 = source_md5.hexdigest()
        if calculated_md5 == md5.replace('"', ""):
            return True
        return False

    def _page_processor(self) -> None:
        proc = multiprocessing.current_process().name
        _logger.debug(f"Running: {proc}")
        self._barrier.wait()
        while True:
            try:
                data = self._download_queue.get(block=True, timeout=0.2)
            except Empty:
                if self._end_event.is_set():
                    break
                continue
            else:
                self._s3_client.download_all_objects(self._bucket_name, data)
                progress_update = {"downloaded": len(data)}
                self._progress_queue.put(progress_update)

    def list_buckets(
        self,
        exclude: list,
        show_excludes: bool,
        show_obsolete: bool,
        only_enabled: bool,
        sort: str,
        sort_revers: bool,
    ) -> None:
        results = []
        s3_buckets = self._s3_client.get_admin_buckets()
        s3ben_buckets = os.listdir(os.path.join(self._backup_root, "active"))
        merged_list = list(dict.fromkeys(s3_buckets + s3ben_buckets))
        for bucket in merged_list:
            bucket_excluded = True if bucket in exclude else ""
            enabled = True if bucket in s3ben_buckets else ""
            obsolete = True if bucket not in s3_buckets else ""
            if not show_excludes and bucket_excluded:
                continue
            if not show_obsolete and obsolete:
                continue
            if only_enabled and not enabled:
                continue
            remote_size = 0
            objects = 0
            unit = ""
            if bucket in s3_buckets:
                bucket_info = self._s3_client.get_bucket(bucket=bucket)
                if "rgw.main" in bucket_info["usage"].keys():
                    original_size = bucket_info["usage"]["rgw.main"].get(
                        "size_utilized"
                    )
                    remote_size = convert_to_human_v2(original_size)
                    original_objects = bucket_info["usage"]["rgw.main"].get(
                        "num_objects"
                    )
                    objects, unit = convert_to_human(original_objects)
            remote_format = ">3d" if isinstance(objects, int) else ">5.2f"
            info = {
                "Bucket": bucket,
                "Owner": bucket_info.get("owner"),
            }
            if not only_enabled:
                info["Enabled"] = enabled
            info.update(
                {
                    "size": original_size,
                    "Remote size": remote_size,
                    "objects": original_objects,
                    "Remote objects": f"{objects:{remote_format}}{unit}",
                }
            )
            if show_excludes and not only_enabled:
                info["Exclude"] = bucket_excluded

            if show_obsolete and not only_enabled:
                info["Obsolete"] = obsolete
            results.append(info)
        if sort == "bucket":
            results = sorted(results, key=lambda k: k["Bucket"], reverse=sort_revers)
        if sort == "owner":
            results = sorted(results, key=lambda k: k["Owner"], reverse=sort_revers)
        if sort == "size":
            results = sorted(results, key=lambda k: k["size"], reverse=sort_revers)
        if sort == "objects":
            results = sorted(results, key=lambda k: k["objects"], reverse=sort_revers)
        for r in results:
            r.pop("size")
            r.pop("objects")
        print(tabulate(results, headers="keys"))

    def cleanup_deleted_items(self, days: int) -> None:
        """
        Method to cleanup deleted items
        :param int days: Days to keep deleted items
        :return: None
        """
        deleted_path = os.path.join(self._backup_root, "deleted")
        date_to_remove = datetime.now() - timedelta(days=days)
        _, dirs, _ = next(os.walk(deleted_path))
        for dir in dirs:
            dir_date = datetime.strptime(dir, "%Y-%m-%d")
            if dir_date > date_to_remove:
                continue
            to_remove = os.path.join(deleted_path, dir)
            _logger.debug(f"Removing {to_remove}")
            shutil.rmtree(path=to_remove)
        _logger.debug("Cleanup done")


class ResolveRemmaping:
    """
    Class to resolve remmapped objects
    :param str backup_root: Path to backup root
    """

    def __init__(self, backup_root: str):
        self._queue = None
        self._remapping_db = os.path.join(backup_root, ".remappings")

    def update_remapping(self, bucket: str, remap: dict) -> None:
        """
        Method to update remapping database
        :param str bucket: bucket name for which remap should be added
        :peram dict remap: dictionary containing remapping information
        :return: None
        """
        b_remaps = {}
        if not os.path.exists(self._remapping_db):
            _logger.warning("Remapping db doesn't exists, creating")
            update = {bucket: remap}
            with open(file=self._remapping_db, mode="w", encoding="utf-8") as f:
                json.dump(update, f)
                return
        with open(file=self._remapping_db, mode="r", encoding="utf-8") as f:
            remappings: dict = json.load(f)
        if bucket in remappings.keys():
            b_remaps = remappings.pop(bucket)
        b_remaps.update(remap)
        remappings.update({bucket: b_remaps})
        with open(file=self._remapping_db, mode="w", encoding="utf-8") as f:
            json.dump(obj=remappings, fp=f)

    def run(self, queue: Queue, event: Event) -> None:
        """
        Method to launch Resolver as a process
        :param multiprocess.Queue queue: multiprocess.Queue class for receiving data
        :param multiprocess.Event event: multiprocess.Event class for receiving end event
        :return: None

        for updating remap db, dictionary must be added to queue:
        {
          "action": "update",
          "data": {
            "bucket": "bucket_name"
            "remap": {
              "object_key_to_match_local_file": "relative/path_to_key"
            }
          }
        }
        """
        _logger.info("starting remapping resolver")
        while not event.is_set():
            try:
                data: dict = queue.get(timeout=1)
            except Empty:
                if event.is_set():
                    break
                continue
            if data.get("action") == "update":
                remap = data.get("data")
                self.update_remapping(**remap)
